import datetime
import logging
from collections.abc import Callable, Iterable, Mapping, Sequence
from decimal import Decimal
from typing import Final

from prettytable import PrettyTable

from .config import config
from .typing import Year

logger = logging.getLogger(__name__)

TAX_YEAR_MONTH: Final = 4
TAX_YEAR_START_DAY: Final = 6
TAX_YEAR_END_DAY: Final = 5


def tax_year_period(tax_year: Year) -> tuple[datetime.date, datetime.date]:
    tax_year_start = datetime.date(tax_year, TAX_YEAR_MONTH, TAX_YEAR_START_DAY)
    tax_year_end = datetime.date(tax_year + 1, TAX_YEAR_MONTH, TAX_YEAR_END_DAY)
    return tax_year_start, tax_year_end


def date_to_tax_year(date: datetime.date) -> Year:
    tax_year_start, _ = tax_year_period(Year(date.year))
    if date >= tax_year_start:
        return Year(tax_year_start.year)
    return Year(tax_year_start.year - 1)


def multifilter(filters: Sequence[Callable] | None, iterable: Iterable) -> Iterable:
    if not filters:
        return iterable
    return filter(lambda x: all(f(x) for f in filters), iterable)


def raise_or_warn(ex: Exception) -> None:
    if config.strict:
        raise ex
    logger.warning(ex)


def read_decimal(val: str, default: Decimal = Decimal("0.0")) -> Decimal:
    return Decimal(val) if val.strip() else default


def dict2str(d: Mapping[str, str]) -> str:
    return str({k: v for k, v in d.items() if v.strip()})


def printtable(
    table: PrettyTable, leading_newline: bool = True, trailing_newline: bool = True
) -> None:
    preamble = epilog = ""

    if leading_newline and config.log_level != logging.CRITICAL:
        preamble = "\n"

    if trailing_newline:
        epilog = "\n"

    print(preamble, table, epilog, sep="")



import os
from collections.abc import Callable, Sequence
from datetime import datetime, timezone
from decimal import Decimal
from pathlib import Path
from typing import Any

import pytest
import yaml

from investir.securitiesdatacache import VERSION, SecuritiesDataCache
from investir.securitiesdataprovider import (
    SecurityData,
    Split,
    YahooFinanceDataProvider,
)
from investir.transaction import Acquisition, Disposal
from investir.trhistory import TrHistory
from investir.typing import ISIN

ORDER1 = Acquisition(
    datetime(2018, 1, 1, tzinfo=timezone.utc),
    isin=ISIN("AMZN-ISIN"),
    name="Amazon",
    amount=Decimal("10.0"),
    quantity=Decimal("1.0"),
)

ORDER2 = Acquisition(
    datetime(2020, 1, 1, tzinfo=timezone.utc),
    isin=ISIN("AMZN-ISIN"),
    name="Amazon",
    amount=Decimal("10.0"),
    quantity=Decimal("1.0"),
)

ORDER3 = Disposal(
    datetime(2022, 1, 1, tzinfo=timezone.utc),
    isin=ISIN("AMZN-ISIN"),
    name="Amazon",
    amount=Decimal("25.0"),
    quantity=Decimal("2.0"),
)

ORDER4 = Acquisition(
    datetime(2024, 1, 1, tzinfo=timezone.utc),
    isin=ISIN("NFLX-ISIN"),
    name="Netflix",
    amount=Decimal("15.0"),
    quantity=Decimal("2.0"),
)

ORDER5 = Acquisition(
    datetime(2024, 3, 1, tzinfo=timezone.utc),
    isin=ISIN("NOTF-ISIN"),
    name="Not Found",
    amount=Decimal("5.0"),
    quantity=Decimal("1.0"),
)

ORDER6 = Acquisition(
    datetime(2024, 6, 1, tzinfo=timezone.utc),
    isin=ISIN("MSFT-ISIN"),
    name="Microsoft",
    amount=Decimal("15.0"),
    quantity=Decimal("2.0"),
)

AMZN_SPLITS = [
    Split(
        date_effective=datetime(2019, 1, 1, tzinfo=timezone.utc), ratio=Decimal("10.0")
    ),
    Split(
        date_effective=datetime(2021, 1, 1, tzinfo=timezone.utc), ratio=Decimal("3.0")
    ),
]

NFLX_SPLITS = [
    Split(
        date_effective=datetime(2023, 1, 1, tzinfo=timezone.utc), ratio=Decimal("5.0")
    ),
]


@pytest.fixture(name="create_securities_data_cache")
def _create_securities_data_cache(mocker) -> Callable:
    def _method(
        tr_hist: TrHistory, cache_file: Path, data: Sequence[SecurityData]
    ) -> tuple[SecuritiesDataCache, Any]:
        data_provider = YahooFinanceDataProvider()
        mock = mocker.patch.object(data_provider, "get_security_data")
        mock.side_effect = data
        return SecuritiesDataCache(data_provider, tr_hist, cache_file), mock

    return _method


def test_securities_data_initialisation_without_cache(
    create_securities_data_cache, tmp_path
):
    cache_file = tmp_path / "cache.yaml"

    tr_hist = TrHistory(orders=[ORDER1, ORDER2, ORDER3, ORDER4, ORDER5])

    data_cache, data_provider_mock = create_securities_data_cache(
        tr_hist,
        cache_file,
        [
            SecurityData(name="Amazon", splits=AMZN_SPLITS),
            SecurityData(name="Netflix", splits=NFLX_SPLITS),
            None,
        ],
    )
    assert data_cache[ISIN("AMZN-ISIN")].name == "Amazon"
    assert data_cache[ISIN("AMZN-ISIN")].splits == AMZN_SPLITS
    assert data_cache[ISIN("NFLX-ISIN")].name == "Netflix"
    assert data_cache[ISIN("NFLX-ISIN")].splits == NFLX_SPLITS
    assert data_cache[ISIN("NOTF-ISIN")].name == "Not Found"
    assert data_cache[ISIN("NOTF-ISIN")].splits == []
    assert data_provider_mock.call_count == 3

    assert cache_file.exists()
    with cache_file.open("r") as file:
        data = yaml.load(file, Loader=yaml.FullLoader)
    assert data["version"] == VERSION
    assert data["securities"].get(ISIN("AMZN-ISIN")).name == "Amazon"
    assert data["securities"].get(ISIN("AMZN-ISIN")).splits == AMZN_SPLITS
    assert data["securities"].get(ISIN("NFLX-ISIN")).name == "Netflix"
    assert data["securities"].get(ISIN("NFLX-ISIN")).splits == NFLX_SPLITS
    assert data["securities"].get(ISIN("NOTF-ISIN")).name == "Not Found"
    assert data["securities"].get(ISIN("NOTF-ISIN")).splits == []

    data_cache, data_provider_mock = create_securities_data_cache(
        tr_hist, cache_file, None
    )
    assert data_cache[ISIN("AMZN-ISIN")].name == "Amazon"
    assert data_cache[ISIN("AMZN-ISIN")].splits == AMZN_SPLITS
    assert data_cache[ISIN("NFLX-ISIN")].name == "Netflix"
    assert data_cache[ISIN("NFLX-ISIN")].splits == NFLX_SPLITS
    assert data_cache[ISIN("NOTF-ISIN")].name == "Not Found"
    assert data_cache[ISIN("NOTF-ISIN")].splits == []
    assert data_provider_mock.call_count == 0


def test_securities_data_cache_is_updated(create_securities_data_cache, tmp_path):
    cache_file = tmp_path / "cache.yaml"
    cache_file.write_text(
        """
    securities:
        AMZN-ISIN: !security
            name: Amazon
            splits:
            - !split '2019-01-01 00:00:00+00:00, 10'
            - !split '2021-01-01 00:00:00+00:00, 3'
            last_updated: 2022-01-01 00:00:00+00:00
        NFLX-ISIN: !security
            name: Netflix
            splits:
            - !split '2021-05-12 00:00:00+00:00, 20'
            last_updated: 2024-01-01 00:00:00+00:00
        """
    )

    amazn_splits = AMZN_SPLITS + [
        Split(date_effective=datetime(2024, 1, 1), ratio=Decimal("40.0"))
    ]

    tr_hist = TrHistory(orders=[ORDER1, ORDER2, ORDER3, ORDER4, ORDER5, ORDER6])

    data_cache, data_provider_mock = create_securities_data_cache(
        tr_hist,
        cache_file,
        [
            SecurityData(name="Amazon", splits=amazn_splits),
            SecurityData(name="Microsoft", splits=[]),
            SecurityData(name="Netflix", splits=NFLX_SPLITS),
            None,
        ],
    )

    assert data_cache[ISIN("AMZN-ISIN")].splits == amazn_splits
    assert data_cache[ISIN("NFLX-ISIN")].splits == NFLX_SPLITS
    assert not data_cache[ISIN("MSFT-ISIN")].splits
    assert data_provider_mock.call_count == 4

    with cache_file.open("r") as file:
        data = yaml.load(file, yaml.FullLoader)

    assert tuple(data["securities"].keys()) == (
        ISIN("AMZN-ISIN"),
        ISIN("MSFT-ISIN"),
        ISIN("NFLX-ISIN"),
        ISIN("NOTF-ISIN"),
    )

    assert data["securities"].get(ISIN("AMZN-ISIN")).name == "Amazon"
    assert data["securities"].get(ISIN("AMZN-ISIN")).splits == amazn_splits
    assert data["securities"].get(ISIN("MSFT-ISIN")).name == "Microsoft"
    assert data["securities"].get(ISIN("MSFT-ISIN")).splits == []
    assert data["securities"].get(ISIN("NFLX-ISIN")).name == "Netflix"
    assert data["securities"].get(ISIN("NFLX-ISIN")).splits == NFLX_SPLITS
    assert data["securities"].get(ISIN("NOTF-ISIN")).name == "Not Found"
    assert data["securities"].get(ISIN("NOTF-ISIN")).splits == []


def test_empty_data_cache_does_not_raise_exception(create_securities_data_cache):
    create_securities_data_cache(
        TrHistory(),
        Path(os.devnull),
        [],
    )



import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import ClassVar, NamedTuple

from .transaction import Dividend, Interest, Order, Transfer

logger = logging.getLogger(__name__)


class ParsingResult(NamedTuple):
    orders: list[Order]
    dividends: list[Dividend]
    transfers: list[Transfer]
    interest: list[Interest]


class Parser(ABC):
    @staticmethod
    @abstractmethod
    def name() -> str:
        pass

    @abstractmethod
    def can_parse(self) -> bool:
        pass

    @abstractmethod
    def parse(self) -> ParsingResult:
        pass


class ParserFactory:
    _parsers: ClassVar[list[type[Parser]]] = []

    @classmethod
    def register_parser(cls, parser: type[Parser]) -> None:
        cls._parsers.append(parser)

    @classmethod
    def create_parser(cls, filename: Path) -> Parser | None:
        for parser_class in cls._parsers:
            parser = parser_class(filename)  # type: ignore[call-arg]
            if parser.can_parse():
                logger.info("Found parser: %s", type(parser).name())
                return parser

        return None



import datetime
from decimal import Decimal

from investir.utils import (
    date_to_tax_year,
    dict2str,
    multifilter,
    read_decimal,
    tax_year_period,
)


def test_tax_year_period():
    tax_year_start, tax_year_end = tax_year_period(2023)
    assert tax_year_start == datetime.date(2023, 4, 6)
    assert tax_year_end == datetime.date(2024, 4, 5)


def test_date_to_tax_year():
    assert date_to_tax_year(datetime.date(2023, 4, 5)) == 2022
    assert date_to_tax_year(datetime.date(2023, 4, 6)) == 2023
    assert date_to_tax_year(datetime.date(2024, 4, 5)) == 2023
    assert date_to_tax_year(datetime.date(2024, 4, 6)) == 2024


def test_multifilter():
    nums = [1, 2, 3, 4, 5, 6]

    nums_filtered = multifilter(None, nums)
    assert nums_filtered == nums

    nums_filtered = multifilter([], nums)
    assert nums_filtered == nums

    nums_filtered = multifilter([lambda x: x % 2 == 0, lambda x: x > 3], nums)
    assert list(nums_filtered) == [4, 6]


def test_read_decimal():
    assert read_decimal("") == Decimal("0.0")
    assert read_decimal(" ") == Decimal("0.0")
    assert read_decimal("0.0") == Decimal("0.0")
    assert read_decimal("2.5") == Decimal("2.5")
    assert read_decimal("", Decimal("2.7")) == Decimal("2.7")


def test_dict2str():
    d = {"a": "A", "b": "", "c": "C", "d": " "}
    assert dict2str(d) == "{'a': 'A', 'c': 'C'}"



import logging
from collections import defaultdict, namedtuple
from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass
from datetime import date, timedelta
from decimal import Decimal
from typing import TypeAlias

from .exceptions import AmbiguousTickerError, IncompleteRecordsError
from .prettytable import PrettyTable
from .securitiesdatacache import SecuritiesDataCache
from .transaction import Acquisition, Disposal, Order
from .trhistory import TrHistory
from .typing import ISIN, Ticker, Year
from .utils import printtable, raise_or_warn

logger = logging.getLogger(__name__)


GroupKey = namedtuple("GroupKey", ["isin", "date", "type"])
GroupDict: TypeAlias = Mapping[GroupKey, Sequence[Order]]


def same_day_match(ord1: Acquisition, ord2: Disposal) -> bool:
    assert ord1.isin == ord2.isin
    return ord1.date == ord2.date


def thirty_days_match(ord1: Acquisition, ord2: Disposal) -> bool:
    assert ord1.isin == ord2.isin
    return ord2.date < ord1.date <= ord2.date + timedelta(days=30)


@dataclass
class CapitalGain:
    disposal: Disposal
    cost: Decimal
    date_acquired: date | None = None

    @property
    def gain_loss(self) -> Decimal:
        # Disposal fees have been added to the `cost` so the gross
        # proceeds are used.
        return self.disposal.amount - self.cost

    @property
    def quantity(self) -> Decimal:
        if self.disposal.original_quantity is not None:
            return self.disposal.original_quantity
        return self.disposal.quantity

    def __str__(self) -> str:
        return (
            f"{self.disposal.date} "
            f"{self.disposal.isin:<4} "
            f"quantity: {self.quantity}, "
            f"cost: £{self.cost:.2f}, proceeds: £{self.disposal.amount}, "
            f"gain: £{self.gain_loss:.2f} "
            f'({self.date_acquired or "Section 104"})'
        )


@dataclass
class Section104Holding:
    quantity: Decimal
    cost: Decimal

    def __init__(self, _date: date, quantity: Decimal, cost: Decimal):
        self.quantity = quantity
        self.cost = cost

    def increase(self, _date: date, quantity: Decimal, cost: Decimal) -> None:
        self.quantity += quantity
        self.cost += cost

    def decrease(self, _date: date, quantity: Decimal, cost: Decimal) -> None:
        self.quantity -= quantity
        self.cost -= cost


class TaxCalculator:
    def __init__(
        self, tr_hist: TrHistory, securities_data: SecuritiesDataCache
    ) -> None:
        self._tr_hist = tr_hist
        self._securities_data = securities_data
        self._acquisitions: dict[ISIN, list[Acquisition]] = defaultdict(list)
        self._disposals: dict[ISIN, list[Disposal]] = defaultdict(list)
        self._holdings: dict[ISIN, Section104Holding] = {}
        self._capital_gains: dict[Year, list[CapitalGain]] = defaultdict(list)

    def capital_gains(self, tax_year: Year | None = None) -> Sequence[CapitalGain]:
        self._calculate_capital_gains()

        if tax_year is not None:
            return self._capital_gains.get(tax_year, [])

        return [cg for cg_group in self._capital_gains.values() for cg in cg_group]

    def holding(self, isin: ISIN) -> Section104Holding | None:
        self._calculate_capital_gains()
        return self._holdings.get(isin)

    def show_capital_gains(  # pylint: disable=too-many-locals
        self,
        tax_year_filter: Year | None,
        ticker_filter: Ticker | None,
        gains_only: bool,
        losses_only: bool,
    ):
        assert not (gains_only and losses_only)

        self._calculate_capital_gains()

        if tax_year_filter is not None:
            tax_years = [tax_year_filter]
        else:
            tax_years = sorted(self._capital_gains.keys())

        for tax_year_idx, tax_year in enumerate(tax_years, 1):
            table = PrettyTable(
                title=f"Tax year {tax_year}-{tax_year + 1}",
                field_names=(
                    "Date Disposed",
                    "Date Acquired",
                    "ISIN",
                    "Name",
                    "Quantity",
                    "Cost (£)",
                    "Proceeds (£)",
                    "Gain/loss (£)",
                ),
            )

            num_disposals = 0
            disposal_proceeds = total_cost = total_gains = total_losses = Decimal("0.0")

            for cg in self.capital_gains(tax_year):
                if ticker_filter is not None and cg.disposal.ticker != ticker_filter:
                    continue

                if gains_only and cg.gain_loss < 0.0:
                    continue

                if losses_only and cg.gain_loss > 0.0:
                    continue

                table.add_row(
                    [
                        cg.disposal.date,
                        cg.date_acquired or "Section 104",
                        cg.disposal.isin,
                        cg.disposal.name,
                        cg.quantity,
                        cg.cost,
                        cg.disposal.amount,
                        cg.gain_loss,
                    ]
                )

                num_disposals += 1
                disposal_proceeds += round(cg.disposal.amount, 2)
                total_cost += round(cg.cost, 2)
                if cg.gain_loss > 0.0:
                    total_gains += cg.gain_loss
                else:
                    total_losses += abs(cg.gain_loss)

            if table.rows:
                printtable(
                    table, leading_newline=tax_year_idx == 1, trailing_newline=False
                )

                def gbp(amount):
                    return "£" + f"{amount:.2f}"

                print(
                    f"{'Number of disposals:':44}{num_disposals:>10}       "
                    f"{'Gains in the year, before losses:':44}{gbp(total_gains):>10}"
                )
                print(
                    f"{'Disposal proceeds:':44}{gbp(disposal_proceeds):>10}       "
                    f"{'Losses in the year:':44}{gbp(total_losses):>10}"
                )
                print(
                    f"{'Allowable costs (including purchase price):':44}"
                    f"{gbp(total_cost):>10}\n"
                )

    def show_holdings(self, ticker_filter: Ticker | None):
        self._calculate_capital_gains()

        table = PrettyTable(
            field_names=[
                "ISIN",
                "Name",
                "Cost (£)",
                "Allocation (%)",
                "Quantity",
                "Average Cost (£)",
            ]
        )

        holdings = []

        if ticker_filter is None:
            holdings = sorted(
                self._holdings.items(), key=lambda x: x[1].cost, reverse=True
            )
        else:
            try:
                isin = self._tr_hist.get_ticker_isin(ticker_filter)
            except AmbiguousTickerError as e:
                logger.warning(e)
            else:
                if isin in self._holdings:
                    holdings = [(isin, self._holdings[isin])]

        total_cost = sum(round(holding.cost, 2) for _, holding in holdings)
        last_idx = len(holdings) - 1

        for idx, (isin, holding) in enumerate(holdings):
            table.add_row(
                [
                    isin,
                    self._tr_hist.get_security_name(isin),
                    holding.cost,
                    holding.cost / total_cost * 100,
                    holding.quantity,
                    holding.cost / holding.quantity,
                ],
                divider=idx == last_idx,
            )

        table.add_row(["", "", total_cost, "", "", ""])

        if holdings:
            printtable(table)

    def _calculate_capital_gains(self) -> None:
        if self._capital_gains or self._holdings:
            # Capital gains already calculated.
            return

        logger.info("Calculating capital gains")

        # First normalise the orders by retroactively adjusting their
        # share quantity for any eventual share sub-division or share
        # consolidation event.
        orders = self._normalise_orders(self._tr_hist.orders)

        # Group together orders that have the same isin, date and type.
        same_day = self._group_same_day(orders)

        for isin, name in self._tr_hist.securities:
            logger.debug("Calculating capital gains for %s (%s)", name, isin)

            # Merge orders that were issued in the same day and have
            # the same type, and place them in the acquisitions or
            # disposals bucket.
            self._merge_same_day(isin, same_day)

            # Match disposed shares with shares acquired in the same
            # day.
            self._match_shares(isin, same_day_match)

            # Match disposed shares with shares acquired up to 30 days
            # past the disposal date.
            self._match_shares(isin, thirty_days_match)

            # Process shares disposed from a Section 104 pool.
            self._process_section104_disposals(isin)

        # Capital gains are calculated ticker by ticker in order to be
        # able to show all the intermediary calculations grouped together
        # as a future improvement. However, for the report, it is more
        # intuitive to show the capital gain transactions ordered by
        # their disposal date.
        for year, events in self._capital_gains.items():
            self._capital_gains[year] = sorted(
                events, key=lambda te: (te.disposal.timestamp, te.disposal.isin)
            )

    def _normalise_orders(self, orders: Sequence[Order]) -> Sequence[Order]:
        return [o.adjust_quantity(self._securities_data[o.isin].splits) for o in orders]

    def _group_same_day(self, orders: Sequence[Order]) -> GroupDict:
        same_day = defaultdict(list)
        for o in orders:
            key = GroupKey(o.isin, o.date, type(o))
            same_day[key].append(o)
        return same_day

    def _merge_same_day(self, isin: ISIN, same_day: GroupDict) -> None:
        security_orders = (
            orders for key, orders in same_day.items() if key.isin == isin
        )

        for orders in security_orders:
            if len(orders) > 1:
                order = Order.merge(*orders)
                logger.debug('    New "same-day" merged order: %s', order)
            else:
                order = orders[0]

            if isinstance(order, Acquisition):
                self._acquisitions[isin].append(order)
            elif isinstance(order, Disposal):
                self._disposals[isin].append(order)

    def _match_shares(
        self, isin: ISIN, match_fn: Callable[[Acquisition, Disposal], bool]
    ):
        acquisits = self._acquisitions[isin]
        disposals = self._disposals[isin]
        matched: set[Order] = set()

        a_idx = 0
        d_idx = 0

        while d_idx < len(disposals):
            if a_idx == len(acquisits):
                a_idx = 0
                d_idx += 1
                continue

            a = acquisits[a_idx]
            d = disposals[d_idx]

            if not match_fn(a, d) or a in matched:
                a_idx += 1
                continue

            matched.add(a)
            matched.add(d)

            if a.quantity > d.quantity:
                a, acquisits[a_idx] = a.split(d.quantity)
                a_idx = 0
                d_idx += 1
            elif d.quantity > a.quantity:
                d, disposals[d_idx] = d.split(a.quantity)
                a_idx += 1
            else:
                a_idx = 0
                d_idx += 1

            self._capital_gains[d.tax_year()].append(
                CapitalGain(d, a.total_cost + d.fees, a.date)
            )

        self._acquisitions[isin] = [o for o in acquisits if o not in matched]
        self._disposals[isin] = [o for o in disposals if o not in matched]

    def _process_section104_disposals(self, isin: ISIN) -> None:
        security_orders = sorted(
            [*self._acquisitions[isin], *self._disposals[isin]],
            key=lambda order: order.date,
        )

        for order in security_orders:
            holding = self._holdings.get(isin)

            if isinstance(order, Acquisition):
                if holding is not None:
                    holding.increase(order.date, order.quantity, order.total_cost)
                else:
                    self._holdings[isin] = Section104Holding(
                        order.date, order.quantity, order.total_cost
                    )
            elif isinstance(order, Disposal):
                if holding is not None:
                    allowable_cost = holding.cost * order.quantity / holding.quantity

                    holding.decrease(order.date, order.quantity, allowable_cost)

                    if holding.quantity < 0.0:
                        raise_or_warn(
                            IncompleteRecordsError(
                                isin, self._tr_hist.get_security_name(isin) or "?"
                            )
                        )
                        logger.warning("Not calculating holding for %s", isin)
                        del self._holdings[isin]
                        break

                    if holding.quantity == Decimal("0.0"):
                        del self._holdings[isin]

                    self._capital_gains[order.tax_year()].append(
                        CapitalGain(order, allowable_cost + order.fees)
                    )
                else:
                    raise_or_warn(
                        IncompleteRecordsError(
                            isin, self._tr_hist.get_security_name(isin) or "?"
                        )
                    )
                    logger.warning("Not calculating holding for %s", isin)
                    del self._holdings[isin]
                    break



import logging
from abc import ABC, abstractmethod
from decimal import Decimal

import yfinance

from .securitydata import SecurityData, Split
from .typing import ISIN

logger = logging.getLogger(__name__)


class SecuritiesDataProvider(ABC):
    @staticmethod
    @abstractmethod
    def name() -> str:
        pass

    @abstractmethod
    def get_security_data(self, isin: ISIN) -> SecurityData | None:
        pass


class NoopDataProvider(SecuritiesDataProvider):
    @staticmethod
    def name() -> str:
        return "Noop"

    def get_security_data(self, _isin: ISIN) -> SecurityData:
        return SecurityData("", [])


class YahooFinanceDataProvider(SecuritiesDataProvider):
    @staticmethod
    def name() -> str:
        return "Yahoo Finance"

    def get_security_data(self, isin: ISIN) -> SecurityData | None:
        try:
            yf_data = yfinance.Ticker(isin)
            name = yf_data.info["shortName"]
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.debug("Exception from yfinance: %s", str(e))
            return None

        splits = [
            Split(pd_date.to_pydatetime(), Decimal(ratio))
            for pd_date, ratio in yf_data.splits.items()
        ]

        return SecurityData(name, splits)



import logging
from pathlib import Path

import yaml

from .securitiesdataprovider import SecuritiesDataProvider
from .securitydata import SecurityData
from .trhistory import TrHistory
from .typing import ISIN

logger = logging.getLogger(__name__)

VERSION = 1


class SecuritiesDataCache:  # pylint: disable=too-few-public-methods
    def __init__(
        self,
        data_provider: SecuritiesDataProvider,
        tr_hist: TrHistory,
        cache_file: Path,
    ) -> None:
        self._data_provider = data_provider
        self._tr_hist = tr_hist
        self._cache_file = cache_file
        self._securities_data: dict[ISIN, SecurityData] = {}

        self._initialise()

    def _initialise(self) -> None:
        self._load_cache()

        orders = self._tr_hist.orders
        update_cache = False

        for isin, name in self._tr_hist.securities:
            if (security_data := self._securities_data.get(isin)) is not None:
                last_order = next(o for o in reversed(orders) if o.isin == isin)

                if security_data.last_updated > last_order.timestamp:
                    logger.debug(
                        "Securities cache for %s (%s) is up-to-date", name, isin
                    )
                    continue

            update_cache = True

            logger.info("Fetching information for %s - %s", isin, name)
            if data := self._data_provider.get_security_data(isin):
                self._securities_data[isin] = data
            else:
                logger.warning("Failed to fetch information for %s - %s", isin, name)
                self._securities_data[isin] = SecurityData(name, [])

        if update_cache:
            self._update_cache()

    def _load_cache(self) -> None:
        if self._cache_file.exists():
            logger.info("Loading securities cache from %s", self._cache_file)

            with self._cache_file.open("r") as file:
                if data := yaml.load(file, Loader=yaml.FullLoader):
                    self._securities_data = data["securities"]

    def _update_cache(self) -> None:
        if self._cache_file.exists():
            logger.info("Updating securities cache on %s", self._cache_file)
        else:
            logger.info("Creating securities cache on %s", self._cache_file)

        self._cache_file.parent.mkdir(parents=True, exist_ok=True)
        securities_info = dict(sorted(self._securities_data.items()))
        data = {"version": VERSION, "securities": securities_info}

        with self._cache_file.open("w") as file:
            yaml.dump(data, file, sort_keys=False)

    def __getitem__(self, isin: ISIN) -> SecurityData:
        return self._securities_data[isin]



from .parser import ParserFactory
from .parsers.freetrade import FreetradeParser
from .parsers.trading212 import Trading212Parser

ParserFactory.register_parser(FreetradeParser)
ParserFactory.register_parser(Trading212Parser)



from collections.abc import Callable
from decimal import Decimal
from typing import Any

import prettytable


class PrettyTable(prettytable.PrettyTable):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        def decimal_fmt(precision: int) -> Callable[[str, Any], str]:
            def _fmt(_field, val) -> str:
                return f"{val:.{precision}f}" if isinstance(val, Decimal) else ""

            return _fmt

        for f in kwargs["field_names"]:
            match f.split()[0], f.split()[-1]:
                case ("Name", _):
                    self.custom_format["Name"] = lambda _, val: f"{val:8}"
                    self.align[f] = "l"

                case ("Quantity", _):
                    self.custom_format["Quantity"] = decimal_fmt(8)
                    self.align[f] = "r"

                case (_, "(£)") | (_, "(%)"):
                    self.custom_format[f] = decimal_fmt(2)
                    self.align[f] = "r"

                case _:
                    self.align[f] = "l"

        self.vrules = prettytable.NONE



import logging
import os
import shlex
from collections.abc import Iterable
from pathlib import Path

import pytest
from typer.testing import CliRunner, Result

from investir.cli import app

PROJECT_DIR = Path(__file__).parent.parent
DATA_FILE = str(PROJECT_DIR / "data" / "freetrade.csv")
OUTPUTS_DIR = PROJECT_DIR / "tests" / "test_cli"
EX_OK = getattr(os, "EX_OK", 0)


@pytest.fixture(name="execute")
def fixture_execute() -> Iterable:
    def _execute(args: list[str]) -> Result:
        runner = CliRunner(mix_stderr=False)
        global_opts = ["--offline", "--cache-file", os.devnull]
        return runner.invoke(app, global_opts + args)

    logging.disable()
    yield _execute
    logging.disable(logging.NOTSET)


test_data = [
    # Orders
    ("orders", "orders"),
    ("orders --tax-year 2022", "orders_ty2022"),
    ("orders --ticker SWKS", "orders_swks"),
    ("orders --acquisitions", "orders_acquisitions"),
    ("orders --disposals", "orders_disposals"),
    (
        "orders --tax-year 2022 --ticker SWKS",
        "orders_ty2022_swks",
    ),
    (
        "orders --tax-year 2022 --acquisitions",
        "orders_ty2022_acquisitions",
    ),
    (
        "orders --tax-year 2022 --disposals",
        "orders_ty2022_disposals",
    ),
    (
        "orders --ticker SWKS --acquisitions",
        "orders_swks_acquisitions",
    ),
    (
        "orders --ticker SWKS --disposals",
        "orders_swks_disposals",
    ),
    (
        "orders --tax-year 2022 --ticker SWKS --acquisitions",
        "orders_ty2022_swks_acquisitions",
    ),
    (
        "orders --tax-year 2022 --ticker SWKS --disposals",
        "orders_ty2022_swks_disposals",
    ),
    # Dividends
    ("dividends", "dividends"),
    ("dividends --tax-year 2022", "dividends_ty2022"),
    ("dividends --ticker AAPL", "dividends_aapl"),
    ("dividends --tax-year 2022 --ticker AAPL", "dividends_ty2022_aapl"),
    # Interest
    ("interest", "interest"),
    ("interest --tax-year 2022", "interest_ty2022"),
    # Transfers
    ("transfers", "transfers"),
    ("transfers --tax-year 2022", "transfers_ty2022"),
    ("transfers --deposits", "transfers_deposits"),
    ("transfers --withdrawals", "transfers_withdrawals"),
    ("transfers --tax-year 2022 --deposits", "transfers_ty2022_deposits"),
    ("transfers --tax-year 2022 --withdrawals", "transfers_ty2022_withdrawals"),
    # Capital gains
    ("capital-gains", "capital_gains"),
    ("capital-gains --tax-year 2022", "capital_gains_ty2022"),
    ("capital-gains --ticker SWKS", "capital_gains_swks"),
    ("capital-gains --ticker AMZN", "capital_gains_amzn"),
    ("capital-gains --gains", "capital_gains_gains"),
    ("capital-gains --losses", "capital_gains_losses"),
    ("capital-gains --ticker SWKS --gains", "capital_gains_swks_gains"),
    ("capital-gains --ticker SWKS --losses", "capital_gains_swks_losses"),
    ("capital-gains --tax-year 2022 --gains", "capital_gains_ty2022_gains"),
    ("capital-gains --tax-year 2022 --losses", "capital_gains_ty2022_losses"),
    (
        "capital-gains --tax-year 2022 --ticker SWKS --gains",
        "capital_gains_ty2022_swks_gains",
    ),
    (
        "capital-gains --tax-year 2022 --ticker SWKS --losses",
        "capital_gains_ty2022_swks_losses",
    ),
    # Holdings
    ("holdings", "holdings"),
    ("holdings --ticker SWKS", "holdings_swks"),
]


@pytest.mark.parametrize("cmd,expected", test_data)
def test_cli_output(request, execute, cmd, expected):
    expected_output = OUTPUTS_DIR / expected
    result = execute(shlex.split(cmd) + [DATA_FILE])

    if request.config.getoption("--regen-outputs"):
        expected_output.write_text(result.stdout)
    else:
        assert not result.stderr
        assert result.exit_code == EX_OK

        assert expected_output.exists()
        assert result.stdout == expected_output.read_text(encoding="utf-8")


def test_capital_gains_multiple_input_files(execute):
    expected_output = OUTPUTS_DIR / "capital_gains"
    data_file1 = str(PROJECT_DIR / "data" / "trading212_2021-2022.csv")
    data_file2 = str(PROJECT_DIR / "data" / "trading212_2022-2023.csv")
    result = execute(["capital-gains", data_file1, data_file2])

    assert not result.stderr
    assert result.exit_code == EX_OK

    assert expected_output.exists()
    assert result.stdout == expected_output.read_text(encoding="utf-8")


def test_orders_command_no_results(execute):
    result = execute(["orders", "--tax-year", "2008", DATA_FILE])
    assert not result.stdout
    assert not result.stderr
    assert result.exit_code == EX_OK


def test_dividends_command_no_results(execute):
    result = execute(["dividends", "--tax-year", "2008", DATA_FILE])
    assert not result.stdout
    assert not result.stderr
    assert result.exit_code == EX_OK


def test_transfers_command_no_results(execute):
    result = execute(["transfers", "--tax-year", "2008", DATA_FILE])
    assert not result.stdout
    assert not result.stderr
    assert result.exit_code == EX_OK


def test_interest_command_no_results(execute):
    result = execute(["interest", "--tax-year", "2008", DATA_FILE])
    assert not result.stdout
    assert not result.stderr
    assert result.exit_code == EX_OK


def test_capital_gains_command_no_results(execute):
    result = execute(["capital-gains", "--tax-year", "2008", DATA_FILE])
    assert not result.stdout
    assert not result.stderr
    assert result.exit_code == EX_OK


def test_holdings_no_results(execute):
    result = execute(["holdings", "--ticker", "NOTFOUND", DATA_FILE])
    assert not result.stdout
    assert not result.stderr
    assert result.exit_code == EX_OK


def test_orders_command_mutually_exclusive_filters(execute):
    result = execute(["orders", "--acquisitions", "--disposals", DATA_FILE])
    assert not result.stdout
    assert "Usage:" in result.stderr
    assert result.exit_code != EX_OK


def test_transfers_command_mutually_exclusive_filters(execute):
    result = execute(["transfers", "--deposits", "--withdrawals", DATA_FILE])
    assert not result.stdout
    assert "Usage:" in result.stderr
    assert result.exit_code != EX_OK


def test_capital_gains_command_mutually_exclusive_filters(execute):
    result = execute(["capital-gains", "--gains", "--losses", DATA_FILE])
    assert not result.stdout
    assert "Usage:" in result.stderr
    assert result.exit_code != EX_OK



from collections.abc import Callable, Sequence
from datetime import datetime, timezone
from decimal import Decimal
from unittest.mock import Mock, PropertyMock

import pandas as pd
import pytest
import yfinance

from investir.securitiesdataprovider import YahooFinanceDataProvider
from investir.securitydata import Split
from investir.typing import ISIN

AMZN_PSERIES = pd.Series(
    data=[10.0, 3.0],
    index=[
        pd.Timestamp(2019, 1, 1, tzinfo=timezone.utc),
        pd.Timestamp(2021, 1, 1, tzinfo=timezone.utc),
    ],
)

NFLX_PSERIES = pd.Series(
    data=[5.0],
    index=[
        pd.Timestamp(2023, 1, 1, tzinfo=timezone.utc),
    ],
)

AMZN_SPLITS = [
    Split(
        date_effective=datetime(2019, 1, 1, tzinfo=timezone.utc), ratio=Decimal("10.0")
    ),
    Split(
        date_effective=datetime(2021, 1, 1, tzinfo=timezone.utc), ratio=Decimal("3.0")
    ),
]

NFLX_SPLITS = [
    Split(
        date_effective=datetime(2023, 1, 1, tzinfo=timezone.utc), ratio=Decimal("5.0")
    ),
]


@pytest.fixture(name="yf_ticker_mocker")
def _yf_ticker_mocker(mocker) -> Callable:
    def _method(
        info_se: Sequence[str] | Exception, splits_se: Sequence[pd.Series]
    ) -> tuple[PropertyMock, PropertyMock]:
        ticker_mock = Mock()

        splits_prop_mock = PropertyMock(side_effect=splits_se)
        type(ticker_mock).splits = splits_prop_mock

        info_prop_mock = PropertyMock(side_effect=info_se)
        type(ticker_mock).info = info_prop_mock

        mocker.patch("yfinance.Ticker", return_value=ticker_mock)
        return info_prop_mock, splits_prop_mock

    return _method


def test_dataprovider_security_data(yf_ticker_mocker):
    info_prop_mock, splits_prop_mock = yf_ticker_mocker(
        [{"shortName": "Amazon"}, {"shortName": "Netflix"}],
        [AMZN_PSERIES, NFLX_PSERIES],
    )
    data_provider = YahooFinanceDataProvider()
    security_data = data_provider.get_security_data(ISIN("AMZN-ISIN"))
    assert security_data.name == "Amazon"
    assert security_data.splits == AMZN_SPLITS
    security_data = data_provider.get_security_data(ISIN("NFLX-ISIN"))
    assert security_data.name == "Netflix"
    assert security_data.splits == NFLX_SPLITS
    assert info_prop_mock.call_count == 2
    assert splits_prop_mock.call_count == 2


def test_dataprovider_security_not_found(yf_ticker_mocker):
    yf_ticker_mocker(yfinance.exceptions.YFException, [])
    data_provider = YahooFinanceDataProvider()
    security_data = data_provider.get_security_data(ISIN("NOT-FOUND"))
    assert security_data is None



from collections.abc import Callable, Mapping, Sequence, ValuesView
from decimal import Decimal
from typing import NamedTuple, TypeVar

from .exceptions import AmbiguousTickerError
from .prettytable import PrettyTable
from .transaction import Acquisition, Dividend, Interest, Order, Transaction, Transfer
from .typing import ISIN, Ticker
from .utils import multifilter, printtable

T = TypeVar("T", bound=Transaction)


def unique_and_sorted(transactions: Sequence[T] | None) -> Sequence[T]:
    """Remove duplicated transactions and sort them by timestamp."""
    return sorted(set(transactions or []), key=lambda tr: tr.timestamp)


class Security(NamedTuple):
    isin: ISIN
    name: str = ""


class TrHistory:
    def __init__(
        self,
        *,
        orders: Sequence[Order] | None = None,
        dividends: Sequence[Dividend] | None = None,
        transfers: Sequence[Transfer] | None = None,
        interest: Sequence[Interest] | None = None,
    ) -> None:
        self._orders = unique_and_sorted(orders)
        self._dividends = unique_and_sorted(dividends)
        self._transfers = unique_and_sorted(transfers)
        self._interest = unique_and_sorted(interest)
        self._securities: Mapping[ISIN, Security] = {}

    @property
    def orders(self) -> Sequence[Order]:
        return self._orders

    @property
    def dividends(self) -> Sequence[Dividend]:
        return self._dividends

    @property
    def transfers(self) -> Sequence[Transfer]:
        return self._transfers

    @property
    def interest(self) -> Sequence[Interest]:
        return self._interest

    @property
    def securities(self) -> ValuesView[Security]:
        return self._securities_map().values()

    def get_security_name(self, isin: ISIN) -> str | None:
        security = self._securities_map().get(isin)
        return security.name if security else None

    def get_ticker_isin(self, ticker: Ticker) -> ISIN | None:
        isins = set(o.isin for o in self._orders if o.ticker == ticker)

        match len(isins):
            case 0:
                return None
            case 1:
                return next(iter(isins))
            case _:
                raise AmbiguousTickerError(ticker)

    def show_orders(self, filters: Sequence[Callable] | None = None) -> None:
        table = PrettyTable(
            field_names=(
                "Date",
                "ISIN",
                "Name",
                "Ticker",
                "Total Cost (£)",
                "Net Proceeds (£)",
                "Quantity",
                "Price (£)",
                "Fees (£)",
            )
        )

        transactions = list(multifilter(filters, self._orders))
        last_idx = len(transactions) - 1
        total_total_cost = total_net_proceeds = total_fees = Decimal("0.0")

        for idx, tr in enumerate(transactions):
            net_proceeds = None
            total_cost = None
            if isinstance(tr, Acquisition):
                total_cost = tr.total_cost
                total_total_cost += round(tr.total_cost, 2)
            else:
                net_proceeds = tr.net_proceeds
                total_net_proceeds += round(tr.net_proceeds, 2)

            divider = (
                idx == last_idx or tr.tax_year() != transactions[idx + 1].tax_year()
            )

            table.add_row(
                [
                    tr.date,
                    tr.isin,
                    tr.name,
                    tr.ticker,
                    total_cost,
                    net_proceeds,
                    tr.quantity,
                    tr.price,
                    tr.fees,
                ],
                divider=divider,
            )

            total_fees += tr.fees

        table.add_row(
            [
                "",
                "",
                "",
                "",
                total_total_cost,
                total_net_proceeds,
                "",
                "",
                total_fees,
            ]
        )

        if transactions:
            printtable(table)

    def show_dividends(self, filters: Sequence[Callable] | None = None) -> None:
        table = PrettyTable(
            field_names=(
                "Date",
                "ISIN",
                "Name",
                "Ticker",
                "Amount (£)",
                "Tax widhheld (£)",
            )
        )

        transactions = list(multifilter(filters, self._dividends))
        last_idx = len(transactions) - 1
        total_paid = total_withheld = Decimal("0.0")

        for idx, tr in enumerate(transactions):
            if tr.withheld is not None:
                total_withheld += round(tr.withheld, 2)

            divider = (
                idx == last_idx or tr.tax_year() != transactions[idx + 1].tax_year()
            )

            table.add_row(
                [tr.date, tr.isin, tr.name, tr.ticker, tr.amount, tr.withheld],
                divider=divider,
            )

            total_paid += tr.amount

        table.add_row(["", "", "", "", total_paid, total_withheld])

        if transactions:
            printtable(table)

    def show_transfers(self, filters: Sequence[Callable] | None = None) -> None:
        table = PrettyTable(field_names=("Date", "Deposited (£)", "Withdrew (£)"))

        transactions = list(multifilter(filters, self._transfers))
        last_idx = len(transactions) - 1
        total_deposited = total_withdrew = Decimal("0.0")

        for idx, tr in enumerate(transactions):
            if tr.amount > 0:
                deposited = tr.amount
                widthdrew = ""
                total_deposited += tr.amount
            else:
                deposited = ""
                widthdrew = abs(tr.amount)
                total_withdrew += abs(tr.amount)

            divider = (
                idx == last_idx or tr.tax_year() != transactions[idx + 1].tax_year()
            )

            table.add_row([tr.date, deposited, widthdrew], divider=divider)

        table.add_row(["", total_deposited, total_withdrew])

        if transactions:
            printtable(table)

    def show_interest(self, filters: Sequence[Callable] | None = None) -> None:
        table = PrettyTable(field_names=("Date", "Amount (£)"))

        transactions = list(multifilter(filters, self._interest))
        last_idx = len(transactions) - 1
        total_interest = Decimal("0.0")

        for idx, tr in enumerate(transactions):
            divider = (
                idx == last_idx or tr.tax_year() != transactions[idx + 1].tax_year()
            )

            table.add_row([tr.date, tr.amount], divider=divider)

            total_interest += tr.amount

        table.add_row(["", total_interest])

        if transactions:
            printtable(table)

    def _securities_map(self) -> Mapping[ISIN, Security]:
        if not self._securities:
            self._securities = {
                o.isin: Security(o.isin, o.name)
                for o in sorted(self._orders, key=lambda o: o.name)
            }
        return self._securities
